unit UglyCodeMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls;

type
  TForm1 = class(TForm)
    Edit1: TEdit;
    Memo1: TMemo;
    btnOpt: TButton;
    Button2: TButton;
    btnUnOpt: TButton;
    Button4: TButton;
    Button5: TButton;
    procedure btnOptClick(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure btnUnOptClick(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure Button5Click(Sender: TObject);
  private
    { Private declarations }
    Testing: Boolean;
  public
    { Public declarations }
    procedure StringCompareOpt( aString: String );
    procedure StringCompareUnOpt( aString: String );
    procedure Log( aString: String );
  end;

  TStringValue = (svA, svFoo, svOne, svHello, svQuestion, svFour);
  TStringValues = Array[TStringValue] of String;
  TStringHash = Array[TStringValue] of Integer;
const
  Strings: TStringValues = ('A', 'Foo', 'One', 'Hello', 'Question', 'Four Score');
  UnmatchedStrings: TStringValues = ('B', 'Fub', 'Two', 'Gdbye', 'Response', 'Hold These');
  StringA = 65;
  StringFoo = 70063;
  StringOne = 79237;
  StringHello = 77838799;
  StringQuestion = -820302029;
  StringFour = -988267591;

var
  Form1: TForm1;

implementation

uses
  Diagnostics,
  TypInfo;
{$R *.dfm}

procedure TForm1.btnOptClick(Sender: TObject);
begin
  StringCompareOpt(Edit1.Text);
end;

procedure TForm1.Button2Click(Sender: TObject);
var
  i: TStringValue;
begin
  for i := Low(TStringValue) to High(TStringValue) do
    Memo1.Lines.Add('String'+Copy(GetEnumName(TypeInfo(TStringValue), Ord(i)),3,255)
    +' = '+Strings[i].GetHashCode.ToString+';');
end;

procedure TForm1.btnUnOptClick(Sender: TObject);
begin
  StringCompareUnOpt(Edit1.Text);
end;

procedure TForm1.Button4Click(Sender: TObject);
var
  i: Integer;
  Watch: TStopwatch;
begin
  Testing := True;
  try
    Watch := TStopwatch.StartNew;
    for i := 0 to 1000000-1 do
    begin
      if Random < 0.5 then
        StringCompareUnOpt(Strings[TStringValue(Random(Ord(High(TStringValue)))+1)])
      else
        StringCompareUnOpt(UnmatchedStrings[TStringValue(Random(Ord(High(TStringValue)))+1)])
    end;
    Watch.Stop;
    Memo1.Lines.Add(Watch.ElapsedMilliseconds.ToString + 'ms');
  finally
    Testing := False;
  end;
end;

procedure TForm1.Button5Click(Sender: TObject);
var
  i: Integer;
  Watch: TStopwatch;
begin
  Testing := True;
  try
    Watch := TStopwatch.StartNew;
    for i := 0 to 1000000-1 do
    begin
      if Random < 0.5 then
        StringCompareOpt(Strings[TStringValue(Random(Ord(High(TStringValue)))+1)])
      else
        StringCompareOpt(UnmatchedStrings[TStringValue(Random(Ord(High(TStringValue)))+1)])
    end;
    Watch.Stop;
    Memo1.Lines.Add(Watch.ElapsedMilliseconds.ToString + 'ms');
  finally
    Testing := False;
  end;
end;

procedure TForm1.Log(aString: String);
begin
  if not Testing then
    Memo1.Lines.Add(aString);
end;

procedure TForm1.StringCompareUnOpt(aString: String);
var
  Found: Boolean;
begin
  Found := False;
  // this code was inspired by early Embarcadero (XE2-4?) FMX TPathData.SetPathString code
  // where it compared a token to strings multiple times even when it was found.
  // Years later they fixed it by using a case statement.
  // It wasn't quite this ugly as it was only comparing the token to 1 character strings
  // I also "improved" this by adding a completely useless Not Found and the
  // catch all unknown
  if aString = 'A' then
  begin
    Log('B');
    Found := True;
  end;
  if (aString = 'One') and (not Found) then
  begin
    Log('Two');
    Found := True;
  end;
  if (aString = 'Foo') and (not Found) then
  begin
    Log('Bar');
    Found := True;
  end;
  if (aString = 'Four Score') and (not Found) then
  begin
    Log('And Seven Years Ago');
    Found := True;
  end;
  if (aString = 'Hello') and (not Found) then
  begin
    Log('World');
    Found := True;
  end;
  if (aString = 'Question') and (not Found) then
  begin
    Log('Answer');
    Found := True;
  end;

  if not Found then
    Log('Unknown');
end;

procedure TForm1.StringCompareOpt(aString: String);
begin
  // "Optimized" string compare
  // Pros:
  //    Shorter looking code (if you disregard setting up the constants)
  //    "Looks" like it could be faster as case is just a integer comparison instead of string comparison
  // Cons:
  //    Wrong (case insensitive vs original case sensitive compare),
  //    Brittle (what if Strings const changes?  Need to regenerate hashcodes.
  //             what is GetHashCode implementation changes?  Wrong answers)
  //    Slower (GetHashCode can be slow)
  case aString.GetHashCode of
    StringA:         Log('B');
    StringOne:       Log('Two');
    StringFoo:       Log('Bar');
    StringFour:      Log('And Seven Years Ago');
    StringHello:     Log('World');
    StringQuestion:  Log('Answer');
  else
    Log('Unknown');
  end;
end;

end.
